/*
    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
    #include <direct.h>

    typedef unsigned char   u_char;
    typedef unsigned short  u_short;
    typedef unsigned int    u_int;
    #define PATHSLASH   '\\'
#else
    #include <unistd.h>
    #include <sys/stat.h>

    #define strnicmp    strncasecmp
    #define stristr     strcasestr
    #define PATHSLASH   '/'
#endif



#define VER     "0.1.1"
#define SIGN    "nexus" "\x01\x00\x00"
#define BUFFSZ  16384



void nexus_seek(FILE *fd, u_int off);
void nexus_read(FILE *fd, void *data, int len);
void nexus_extract(FILE *fd, u_char *fname, int len);
void nexus_dec(u_char *data, int len);
void nexus_alternative_extract(FILE *fd, u_char *fname, u_char method);
void create_dir(u_char *name);
void std_err(void);



u_char  xor;



int main(int argc, char *argv[]) {
    FILE    *fd;
    u_int   i,
            tmp,
            off,
            nameoff,
            diroff,
            totlen,
            tot,
            len;
    int     listonly = 0,
            decode_only = 0;
    u_short nlen;
    u_char  type,
            boh,
            *fname,
            *buff,
            *folder = ".",
            *pattern = NULL;

    setbuf(stdout, NULL);

    fputs("\n"
        "Nexus files extractor "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <file.DAT>\n"
            "\n"
            "Options:\n"
            "-l       list files without extracting them\n"
            "-d DIR   the folder where the files will be extracted (default %s)\n"
            "-p PAT   extracts only the files containing PAT, like -p .ogg for example\n"
            "-x       decodes the entire file without extracting the data (useless)\n"
            "\n",
            argv[0], folder);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'l': listonly = 1;         break;
            case 'd': folder   = argv[++i]; break;
            case 'p': pattern  = argv[++i]; break;
            case 'x': decode_only = 1;      break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                } break;
        }
    }

    printf("- open file %s\n", argv[argc]);
    fd = fopen(argv[argc], "rb");
    if(!fd) std_err();

    printf("- enter in folder %s\n", folder);
    if(chdir(folder) < 0) std_err();

    buff = malloc(0xffff + 1);
    if(!buff) std_err();

    xor = 0x2a;
    nexus_read(fd, buff, 8);

    if(memcmp(buff, SIGN, sizeof(SIGN) - 1)) {
        fputs("- movie files decryption enabled\n", stdout);
        sprintf(buff, "%s.bik", argv[argc]);
        printf("- output file: %s\n", buff);
        nexus_alternative_extract(fd, buff, 1);
        fclose(fd);
        return(0);
    }

    if(decode_only) {
        sprintf(buff, "%s.decoded", argv[argc]);
        printf("- output file: %s\n", buff);
        nexus_alternative_extract(fd, buff, 0);
        return(0);
    }

    nexus_read(fd, &tmp, 4);            // zero
    nexus_read(fd, &totlen, 4);         // total package length
    nexus_read(fd, &tmp, 4);            // don't know
    nexus_read(fd, &boh, 1);            // don't know

    nameoff = ftell(fd);

    for(tot = 0; !feof(fd); tot++) {
        nexus_read(fd, &type, 1);       // 0 for files, 1 for folders
        nexus_read(fd, &nlen, 2);       // length of the filename
        nexus_read(fd, buff, nlen);     // filename

        if(listonly) {
            buff[nlen] = 0;
            if(!pattern || (pattern && stristr(buff, pattern))) {
                printf("  %s\n", buff);
            }
        }

        if(type) {
            nexus_read(fd, &tmp, 4);    // offset
            nexus_read(fd, &tmp, 4);    // don't know (length???)
            nexus_read(fd, &tmp, 4);    // don't know
            off = len = 0;
        } else {
            nexus_read(fd, &off, 4);    // offset
            nexus_read(fd, &len, 4);    // length
        }

        nexus_read(fd, &boh, 1);        // don't know

        if((off + len) >= totlen) break;
    }

    if(listonly) {
        printf("- %u files listed\n\n", tot);
        return(0);
    }

    diroff = ftell(fd);
    fname  = buff;

    for(i = 0; i < tot; i++) {
        nexus_seek(fd, nameoff);

        nexus_read(fd, &type, 1);
        nexus_read(fd, &nlen, 2);

        if(type) {
            nexus_read(fd, buff, nlen);
            buff[nlen++] = PATHSLASH;
            buff[nlen] = 0;
            fname = buff + nlen;

            nexus_read(fd, &tmp, 4);
            nexus_read(fd, &tmp, 4);
            nexus_read(fd, &tmp, 4);
        } else {
            nexus_read(fd, fname, nlen);
            fname[nlen] = 0;
            nexus_read(fd, &off, 4);
            nexus_read(fd, &len, 4);
        }

        nexus_read(fd, &boh, 1);

        nameoff = ftell(fd);

        if(type) continue;
        if(pattern && !stristr(fname, pattern)) continue;

        create_dir(buff);
        printf("  %s\n", buff);

        nexus_seek(fd, diroff + off);
        nexus_extract(fd, buff, len);
    }

    fclose(fd);
    printf("- %u files extracted\n\n", i);
    return(0);
}



void nexus_seek(FILE *fd, u_int off) {
    fseek(fd, off, SEEK_SET);
    xor = 0x2a + (off * 0x55);
}



void nexus_read(FILE *fd, void *data, int len) {
    fread(data, len, 1, fd);
    nexus_dec(data, len);
}



void nexus_extract(FILE *fd, u_char *fname, int len) {
    FILE    *fdo;
    int     rlen;
    static u_char   buff[BUFFSZ];

    fdo = fopen(fname, "wb");
    if(!fdo) std_err();

    for(rlen = sizeof(buff); len; len -= rlen) {
        if(len < rlen) rlen = len;
        fread(buff, rlen, 1, fd);
        nexus_dec(buff, rlen);
        fwrite(buff, rlen, 1, fdo);
    }

    fclose(fdo);
}



void nexus_dec(u_char *data, int len) {
    u_char  *limit;

    for(limit = data + len; data < limit; data++) {
        *data ^= xor;
        xor += 0x55;
    }
}



void nexus_alternative_extract(FILE *fd, u_char *fname, u_char method) {
    FILE    *fdo;
    int     len;
    u_char  *p,
            *l;
    static u_char   buff[BUFFSZ];

    fdo = fopen(fname, "wb");
    if(!fdo) std_err();

    fseek(fd, 0, SEEK_SET);

    if(method) {
        len = fread(buff, 1, 0x1092, fd);
        xor = buff[0];
//        fread(&xor, 1, 1, fd);
        if(xor == 'B') {
            printf("\n  The file is already not encrypted!\n\n");
            exit(1);
        }
//        fwrite("B", 1, 1, fdo); // yes here is B for BIK files!
        method = 0xAA;
        buff[0] = 'B';
        l = buff + len;
        for(p = buff + 1; p < l; p++) {
            *p ^= xor;
            xor += method;
        }
        fwrite(buff, 1, len, fdo);
        xor    = 0;     // plain text
        method = 0;
    } else {
        xor    = 0x2a;
        method = 0x55;
    }

    while((len = fread(buff, 1, sizeof(buff), fd))) {

        l = buff + len;
        for(p = buff; p < l; p++) {
            *p ^= xor;
            xor += method;
        }

        fwrite(buff, len, 1, fdo);
    }

    fclose(fdo);
}



void create_dir(u_char *name) {
    u_char  *p,
            *l;

    p = name;
    for(;;) {
        l = strchr(p, '\\');
        if(!l) {
            l = strchr(p, '/');
            if(!l) break;
        }
        *l = 0;
        p = l + 1;
#ifdef WIN32
        mkdir(name);
#else
        mkdir(name, 0755);
#endif
        *l = PATHSLASH;
    }
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


